module RoundRobin where

import Control.Monad
import Data.List

crossMatch :: [a] -> [a] -> (a -> a -> IO b) -> IO [[b]]
crossMatch xs ys match = mapM f xs
  where
    f x = mapM (match x) ys

crossStringTable :: [String] -> [String] -> [[a]] -> (a -> String) -> [[String]]
crossStringTable xns yns objs sho = guideLine : zipWith (:) xns sobjs
  where
    sobjs = map (map sho) objs
    guideLine = "" : map (' ':) yns

roundRobin :: [a] -> (a -> a -> IO b) -> IO [[b]]
roundRobin xs match = mapM f $ zip xs (inits xs)
  where
    f (x, rs) = mapM (match x) rs

completeTable :: a -> (a -> a) -> [[a]] -> [[a]]
completeTable defval neg objs = zipWith (++) objs $ map ((defval:) . map neg) $ transpose objs ++ [[]]

makeStringTable :: [String] -> [String] -> [[a]] -> (a -> String) -> [String] -> ([a] -> [String]) -> [[String]]
makeStringTable rownames colnames objs sho place total = guideLine : zipWith (:) (map (' ':) rownames) sobjs
  where
    guideLine = "" : map (' ':) colnames ++ [""] ++ place
    sobjs = stringize $ addTotal objs
    addTotal = map f
      where
        f line = (" " : total line, line)
    stringize os = map (\(hd, r) -> map f r ++ hd) os
      where
        f v = ' ':sho v

renderAsText :: [[String]] -> String
renderAsText table = unlines $ map concat $ fill table
  where
    fill lines = transpose $ map fillColumn $ transpose lines 
    fillColumn col = map (pad w) col
      where
        w = maximum $ map width col
    width s = sum $ map f s
      where
        f c | c < '\256' = 1
          | otherwise = 2
    pad n s = replicate (n - width s) ' ' ++ s
