module Swords where

import Char
import Game
import qualified Bag as B
import GameUtil
import {-# SOURCE #-} PublicStorage
import qualified CharRef as CR

import Control.Monad
import Data.List
import qualified Data.Map as M

constantSwords :: [Sword]
constantSwords =
  [ slash, rapid, heavy, shock, poison, absolute, magical, death, heat, protect, accel, shield, mirror, life
  , nuru, spark, firecracker, eruption, devour, disaster, perish, finale, combo, wave, dance, scatter, toxic
  , abortion, destroy, battle, confusion, bloodsuck, robbery, dark, antimagic, overweight, solid, sonic
  , light, cool, decay, decel, gentle, flashfire, alert, evasion, escutcheon, amnion, barrier, soul, push
  , drive, wriggle, attraction, first, second, communication, chant, prayer, prospect, dragonize, limit
  , suppression, vibration, toxicant, diffuse, train, fictional, multi, replica, reverb, echo, Swords.bubble
  , spirit, fetal, sawhair, sawa, purple, holand, servant, leg, dog, ko, yoshi, holy, geki, emperor, bone
  , princess, summon, age
  ]

variableSwordsPositive :: [Int -> Sword]
variableSwordsPositive = [recovery, rage, heal]

variableSwordsNatural :: [Int -> Sword]
variableSwordsNatural = [ritual, delay, sheathe, derive, recurrence, resonance]

m100Swords :: [Sword]
m100Swords = constantSwords
  ++ liftM2 ($) variableSwordsPositive [1..20]
  ++ liftM2 ($) variableSwordsNatural [0..20]

m2011aSpecificSwords :: [Sword]
m2011aSpecificSwords =
  [ recoveryA, perishA, rageA, fang, two, three, four, five, six, seven, abortionA, coolA, decayA, decelA
  , darkA, antimagicA, attractionA, pushA, healA, chantA, rogation, ritualA, blaze, deriveA, slow, danger
  , recurrenceA, resonanceA, ageA, mkbubble, mkfence, mkgate, mkwall, mkice, mkmist, mktear
  ]

allSwords :: [Sword]
allSwords = m100Swords ++ m2011aSpecificSwords ++ map victory [1..20]

-- classic
slash = (sw "斬撃剣" "斬" $ attack $ ATK - DEF)
rapid = (sw "高速剣" "高" $ attack $ ATK - DEF) { swTurn = 0, swLimited = True }
heavy = (sw "重撃剣" "重" $ attack $ 2 * ATK - DEF) { swTurn = 4 }
shock = (sw "衝撃剣" "衝" $ attackWith (ATK - DEF) putStun) { swTurn = 2 }
poison = (sw "蟲毒剣" "毒" $ attackWith (ATK - DEF) putPoison) { swLimited = True }
absolute = (sw "絶対剣" "絶" $ attack ATK) { swTurn = 3 }
magical = (sw "魔法剣" "魔" $ attack $ dfdiv (HP + 1) 2) { swTurn = 2, swLimited = True }
death = (sw "デス剣" "死" $ attack HP) { swTurn = 6, swLimited = True }
heat = (sw "加熱剣" "熱" $ gain "攻撃力" csAtk $ \v -> v{ csAtk = csAtk v + 1 })
protect = (sw "加護剣" "護" $ gain "防御力" csDef $ \v -> v{ csDef = csDef v + 1 })
accel = (sw "加速剣" "速" $ gain "素早さ" csSpd $ \v -> v{ csSpd = csSpd v + 1})
shield = (sw "盾の剣" "盾" $ putGuard Guard)
mirror = (sw "鏡の剣" "鏡" $ putDefence Reflect) { swTurn = 2, swLimited = True }
life = (sw "命の剣" "命" $ putRevive Life) { swLimited = True }
recovery n = (swn n "回復剣" "回" $ aux $ \_ ec -> recoverAll (ecOwner ec))

-- m100
nuru = (sw "ヌル剣" "ヌ" $ attack 0) { swTurn = 0 } -- named 'nuru' so that it won't be confused with Prelude.null
spark = (sw "咲火剣" "咲" $ attack 1)
firecracker = (sw "炸撃剣" "炸" $ attack 1)
eruption = (sw "爆炎剣" "爆" $ attack 4) { swMP = Just 1 }
devour = (sw "貪炎剣" "貪" $ attackAnd 6 $ const recoverSelf) { swMP = Just 5 }
disaster = (sw "災厄剣" "災" $ EAtk [simpleRE $ simpleAttackPrim 9, simpleRE $ attackPrim defaultAS{ asSelf = True } 9]) { swMP = Just 5 }
perish = (sw "滅奏剣" "滅" $ attack $ dfpow 2 TURN) { swMP = Just 5 }
finale = (sw "幕引剣" "幕" $ attack 100) { swTurn = 20, swLimited = True }
combo = (sw "連撃剣" "連" $ attackTimes 2 $ ATK - DEF) { swTurn = 3 }
wave = (sw "波状剣" "波" $ attackTimes 5 $ ATK - DEF) { swTurn = 6 }
dance = (sw "舞踏剣" "舞" $ attackThenUnuse $ ATK - DEF) { swTurn = 3 }
scatter = (sw "全体剣" "全" $ EAtk [simpleRE $ attackPrim defaultAS{ asAll = True } $ ATK - DEF]) { swTurn = 2 }
toxic = (sw "猛毒剣" "猛" $ attackWith (ATK - DEF) $ putBad "猛毒" sbBadPoison $ \b -> b{ sbBadPoison = sbBadPoison b + 1 }) { swTurn = 2, swLimited = True }
abortion = (sw "堕胎剣" "堕" $ attackWith (ATK - DEF) putAbortion) { swTurn = 2, swLimited = True }
destroy = (sw "破壊剣" "壊" $ attackWith (ATK - DEF) destroySword) { swLimited = True }
battle = (sw "剣戟剣" "戟" $ attackWith (ATK - DEF) pressurize) { swTurn = 2 }
confusion = (sw "撹乱剣" "撹" $ attackWith (ATK - DEF) reverseSwords) { swLimited = True }
bloodsuck = (sw "吸血剣" "吸" $ attackAnd (ATK - DEF) $ const recoverSelf) { swLimited = True }
robbery = (sw "強盗剣" "盗" $ attackAnd (ATK - DEF) stealSword) { swTurn = 2, swLimited = True }
dark = (sw "闇黒剣" "闇" $ attackAS (ATK - DEF) defaultAS{ asPenetrate = True }) { swMP = Just 2 }
antimagic = (sw "魔封剣" "封" $ attackAS (ATK - DEF) $ defaultAS{ asMpDamage = True })
overweight = (sw "過重剣" "過" $ attack $ 3 * ATK - DEF) { swTurn = 5 }
solid = (sw "確実剣" "確" $ attack $ dfdiv ATK 2)
sonic = (sw "音速剣" "音" $ attackTimes 10 $ dfdiv ATK 5 - DEF) { swTurn = 0, swLimited = True }
light = (sw "光速剣" "光" $ attackTimes 100 $ dfdiv ATK 10 - DEF) { swTurn = 0, swLimited = True }
rage n = (swn n "激怒剣" "怒" $ attack $ SelfMaxHP - SelfHP)
cool = (sw "冷却剣" "冷" $ silentAttack resetAttack) { swMP = Just 1 }
decay = (sw "剥落剣" "剥" $ silentAttack resetDefence) { swMP = Just 1 }
decel = (sw "減速剣" "減" $ silentAttack resetSpeed) { swMP = Just 1 }
gentle = (sw "優撃剣" "優" $ silentAttack recoverAll) { swTurn = 2, swLimited = True }
flashfire = (sw "炎上剣" "炎" flashfireAttack) { swMP = Just 5 }
alert = (sw "警戒剣" "警" $ putGuard Guard) { swTurn = 0, swLimited = True }
evasion = (sw "回避剣" "避" $ putDefence Detect) { swLimited = True }
escutcheon = (sw "楯の剣" "楯" $ putGuard Shield) { swMP = Just 1 }
amnion = (sw "羊膜剣" "膜" $ putGuard Cusion) { swMP = Just 2 }
barrier = (sw "障壁剣" "壁" $ putGuard Barrier) { swMP = Just 3 }
soul = (sw "魂の剣" "魂" $ putRevive Soul) { swTurn = 3, swLimited = True }
push = (sw "押分剣" "押" putPush) { swTurn = 0 }
drive = (sw "駆動剣" "駆" $ putOther "ドライブ" soDrive $ \so -> so{ soDrive = soDrive so + 1 })
wriggle = (sw "蠕動剣" "蠕" $ putGuard Flip)
attraction = (sw "誘引剣" "誘" putProvoke) { swLimited = True }
first = (sw "先行剣" "先" $ putOther "クイック" soQuick $ \so -> so{ soQuick = soQuick so + 1 }) { swTurn = 0, swLimited = True }
second = (sw "後続剣" "後" $ putOther "スロー" soSlow $ \so -> so{ soSlow = soSlow so + 1 }) { swTurn = 0, swLimited = True }
communication = (sw "疎通剣" "疎" $ putOtherOne "ヘヴンズドア" soHeavensDoor $ \so -> so{ soHeavensDoor = True }) { swTurn = 0, swLimited = True }
heal n = (swn n "快癒剣" "癒" $ aux $ \_ _ -> removeAllBad)
chant = (sw "詠唱剣" "詠" $ gainMp 1)
prayer = (sw "黙祷剣" "黙" $ gainMp 3) { swLimited = True }
ritual n = (swn n "儀式剣" "儀" $ gain "AP" csAp $ \v -> v{ csAp = csAp v + 3})
prospect = (sw "臥竜剣" "臥" gainProspect) { swTurn = 0, swLimited = True }
dragonize = (sw "竜化剣" "竜" $ aux $ \_ _ -> doubleStatus) { swTurn = 10, swLimited = True }
limit = (sw "限定剣" "限" $ ctl "リミテッドを得た" (const True) $ \ss -> Just ss{ ssLimitAdded = True }) { swTurn = 0, swLimited = True }
suppression = (sw "抑制剣" "抑" $ ctl "使用済みになった" (\(cs, _) -> csHp cs * 2 >= csMaxHp cs) $ \ss -> Just ss{ ssUsed = True }) { swTurn = 0 }
delay n = (swpn n "遅延剣" "遅" $ ctl "使用済みになった" (const True) $ \ss -> Just ss{ ssUsed = True })
sheathe n = (swn n "収鞘剣" "鞘" $ ctl "消滅した" (const True) $ const Nothing)
derive n = (swn n "導来剣" "導" putEarly)
vibration = (sw "振動剣" "振" $ ctl "スタン与えを得た" (isAtk . snd) $ \ss -> Just ss{ ssStun = ssStun ss + 1 }) { swLimited = True }
toxicant = (sw "毒薬剣" "薬" $ ctl "毒与えを得た" (isAtk . snd) $ \ss -> Just ss{ ssPoison = ssPoison ss + 1 }) { swLimited = True }
diffuse = (sw "拡散剣" "散" $ ctl "全体化を得た" (liftM2 (||) isAtk isAux . snd) $ \ss -> Just ss{ ssAll = True }) { swLimited = True }
train = (sw "鍛錬剣" "鍛" $ ctlAll "重撃剣になった" ((==slash) . snd) $ Just . changeSwordType (initialSS heavy)) { swTurn = 4, swLimited = True }
recurrence n = (swn n "再来剣" "再" makeAllUnused)
resonance n = (swn n "共鳴剣" "鳴" ctlResonance)
fictional = (sw "虚構剣" "虚" $ EAdd $ \gs ec -> Just $ initialSS nuru)
multi = (sw "増殖剣" "増" $ EAdd $ \gs ec -> nextSwordAsSubtarget gs ec) { swMP = Just 1 }
replica = (sw "複製剣" "複" $ EAdd $ \gs ec -> enemySword gs ec) { swLimited = True }
reverb = (sw "残響剣" "残" $ EBld $ primaryChar id) { swTurn = 0, swLimited = True, swMP = Just 5 }
echo = (sw "反響剣" "反" $ EBld $ primaryChar opponent) { swTurn = 0, swLimited = True, swMP = Just 5 }
bubble = (sw "泡溢剣" "泡" $ bld bubblia) { swTurn = 0, swMP = Just 1 }
spirit = (sw "精霊剣" "精" $ bld wisp) { swTurn = 0, swMP = Just 4 }
fetal = (sw "胎胎剣" "胎" $ bld matrox) { swTurn = 0, swLimited = True, swMP = Just 3 }
sawhair = (sw "鋸髪剣" "鋸" $ bld deathEliyna) { swTurn = 0, swLimited = True, swMP = Just 5 }
sawa = (sw "相沢剣" "沢" $ bld aizawa) { swTurn = 0, swLimited = True, swMP = Just 5 }
purple = (sw "紫電剣" "紫" $ bld shiden) { swTurn = 0, swLimited = True, swMP = Just 5 }
holand = (sw "尾蘭剣" "蘭" $ bld thoughtlessHoland) { swTurn = 0, swLimited = True, swMP = Just 5 }
servant = (sw "従僕剣" "従" $ bld charles) { swTurn = 0, swLimited = True, swMP = Just 5 }
leg = (sw "弩脚剣" "脚" $ bld rushiekun) { swTurn = 0, swLimited = True, swMP = Just 5 }
dog = (sw "狗咬剣" "狗" $ bld wanderingRaika) { swTurn = 0, swLimited = True, swMP = Just 5 }
ko = (sw "ｺｳ速剣" "ｺｳ" $ bld kou) { swTurn = 0, swLimited = True, swMP = Just 5 }
yoshi = (sw "吉宏剣" "吉" $ bld takemotoYoshihiro) { swTurn = 0, swLimited = True, swMP = Just 5 }
holy = (sw "聖重剣" "聖" $ bld zoara) { swTurn = 0, swLimited = True, swMP = Just 6 }
geki = (sw "撃鉄剣" "撃" $ bld gekitetsu) { swTurn = 0, swLimited = True, swMP = Just 5 }
emperor = (sw "鶏皇剣" "鶏" $ bld gekitetsuAct6) { swTurn = 0, swLimited = True, swMP = Just 13 }
bone = (sw "怪骨剣" "怪" $ bld meteora) { swTurn = 0, swLimited = True, swMP = Just 7 }
princess = (sw "歪姫剣" "歪" $ bld shian) { swTurn = 0, swLimited = True, swMP = Just 5 }
summon = (sw "招龍剣" "龍" $ bld dragon) { swTurn = 0, swLimited = True, swMP = Just 10 }
age = (sw "星霜剣" "星" $ EEnv $ \_ _ -> incrementTurnCount) { swMP = Just 1 }

-- m2011a
recoveryA = (swa "回復剣" "回" $ aux $ \_ ec -> recoverAll (ecOwner ec)) { swReactsToDanger = True, swLimited = True }
perishA = (swa "滅奏剣" "滅" $ attack $ dfpow 2 TURN) { swReactsToDanger = True, swLimited = True }
rageA = (swa "激怒剣" "怒" $ attack $ SelfMaxHP - SelfHP) { swReactsToDanger = True, swLimited = True }
fang = (swa "牙撃剣" "牙" $ attack $ 2 * ATK - DEF) { swTurn = 2 }
two = (swa "二撃剣" "二" $ attackTimes 2 $ ATK - DEF) { swTurn = 2 }
three = (swa "三撃剣" "三" $ attackTimes 3 $ ATK - DEF) { swTurn = 3 }
four = (swa "四撃剣" "四" $ attackTimes 4 $ ATK - DEF) { swTurn = 4 }
five = (swa "五撃剣" "五" $ attackTimes 5 $ ATK - DEF) { swTurn = 5 }
six = (swa "六撃剣" "六" $ attackTimes 6 $ ATK - DEF) { swTurn = 6 }
seven = (swa "七撃剣" "七" $ attackTimes 7 $ ATK - DEF) { swTurn = 7 }
abortionA = (swa "堕胎剣" "堕" $ attackWith (ATK - DEF) putAbortion) { swLimited = True }
coolA = (swa "冷却剣" "冷" $ attackWith (ATK - DEF) $ const resetAttack) { swTurn = 2 }
decayA = (swa "剥落剣" "剥" $ attackWith (ATK - DEF) $ const resetDefence) { swTurn = 2 }
decelA = (swa "減速剣" "減" $ attackWith (ATK - DEF) $ const resetSpeed) { swTurn = 2 }
darkA = (swa "闇黒剣" "闇" $ attackAS (ATK - DEF) defaultAS{ asPenetrate = True }) { swTurn = 2 }
antimagicA = (swa "魔封剣" "封" $ attackAS (ATK - DEF) $ defaultAS{ asMpDamage = True }) { swLimited = True }
pushA = (swa "押分剣" "押" putPush) { swTurn = 0, swLimited = True }
attractionA = (swa "誘引剣" "誘" putProvoke) { swTurn = 0, swLimited = True }
healA = (swa "快癒剣" "癒" $ aux $ \_ _ -> removeAllBad) { swReactsToIllness = True, swLimited = True }
chantA = (swa "詠唱剣" "詠" $ gainMp 1) { swTurn = 0 }
rogation = (swa "祈祷剣" "祈" $ gainMp 2) { swTurn = 0, swLimited = True }
ritualA = (swa "儀式剣" "儀" $ gain "予備AP" csReservedAp $ \v -> v{ csReservedAp = csReservedAp v + 3 }) { swTurn = 0, swLimited = True }
blaze = (swa "灼熱剣" "灼" $ aux $ \_ _ -> doubleAttack) { swTurn = 3, swLimited = True }
deriveA = (swa "導来剣" "導" putEarly) { swTurn = 0, swLimited = True }
slow = (swa "遅滞剣" "遅" $ ctl "遅滞起動を得た" (const True) $ \ss -> Just $ ss{ ssEarly = ssEarly ss - 1 }) { swTurn = 0, swLimited = True }
danger = (swa "危殆剣" "危" $ ctl "危殆起動を得た" (const True) $ \ss -> Just $ ss{ ssForcedReactsToDanger = True }) { swTurn = 0, swLimited = True }
recurrenceA = (swa "再来剣" "再" makeAllUnused) { swTurn = 3, swLimited = True }
resonanceA = (swa "共鳴剣" "鳴" ctlResonance) { swTurn = 4, swLimited = True }
ageA = (swa "星霜剣" "星" $ EEnv $ \_ _ -> incrementTurnCount)
mkbubble = (swa "造泡剣" "泡" $ bld PublicStorage.bubble) { swTurn = 0, swMP = Just 2, swLimited = True }
mkfence = (swa "造柵剣" "柵" $ bld fence) { swTurn = 0, swMP = Just 3, swLimited = True }
mkgate = (swa "造門剣" "門" $ bld gate) { swTurn = 0, swMP = Just 4, swLimited = True }
mkwall = (swa "造壁剣" "壁" $ bld wall) { swTurn = 0, swMP = Just 5, swLimited = True }
mkice = (swa "造氷剣" "氷" $ bld icia) { swTurn = 0, swMP = Just 3, swLimited = True }
mkmist = (swa "造霧剣" "霧" $ bld mistia) { swTurn = 0, swMP = Just 4, swLimited = True }
mktear = (swa "造涙剣" "涙" $ bld undine) { swTurn = 0, swMP = Just 5, swLimited = True }

-- private
victory n = (swn n "勝利剣" "勝" $ aux $ \_ _ _ -> return ()) { swNamespace = Private }

data AttackStyle = AS
  { asMpDamage :: Bool
  , asAddition :: GameState -> CharRef -> CharRef -> Game ()
  , asNext :: GameState -> Integer -> CharRef -> CharRef -> Game [CharRef]
  , asPenetrate :: Bool
  , asAll :: Bool
  , asSelf :: Bool
  }

defaultAS :: AttackStyle
defaultAS = AS
  { asMpDamage = False
  , asAddition = \gs x y -> return ()
  , asNext = \_ _ _ _ -> return []
  , asPenetrate = False
  , asAll = False
  , asSelf = False
  }

attackPrim :: AttackStyle -> DamageFormula -> PrimEffect
attackPrim as df
  | asSelf as = targetNone $ \ec -> gen (ecOwner ec) ec
  | otherwise = TargetEnemy{ teDefendable = not $ asPenetrate as, teAll = asAll as, teGen = gen }
  where
    gen y ec = AC $ \gs -> do
      let x = ecOwner ec
      val <- handleDriveAndGuard (asPenetrate as) x y $ evalDamageFormulaFromGS gs x y df
      return ([], Just $ atk x y val)
    atk x y val = AC $ \gs -> do
      atkDamage (asMpDamage as) x y val
      when (val > 0) $ asAddition as gs x y
      return (if val > 0 then [y] else [], Just $ doNext val x y)
    doNext k x y = AC $ \gs -> do
      damaged <- asNext as gs k x y
      return (damaged, Nothing)

simpleAttackPrim :: DamageFormula -> PrimEffect
simpleAttackPrim = attackPrim defaultAS

attackTimes :: Int -> DamageFormula -> Effect
attackTimes n df = EAtk [RE{ reRepeat = n, rePrim = simpleAttackPrim df }]

attackAS :: DamageFormula -> AttackStyle -> Effect
attackAS df as = EAtk [simpleRE $ attackPrim as df]

attack :: DamageFormula -> Effect
attack df = attackAS df defaultAS

attackWith :: DamageFormula -> (GameState -> CharRef -> CharRef -> Game ()) -> Effect
attackWith df add = attackAS df defaultAS{ asAddition = add }

attackAnd :: DamageFormula -> (GameState -> Integer -> CharRef -> CharRef -> Game [CharRef]) -> Effect
attackAnd df next = attackAS df defaultAS{ asNext = next }

chaseSword :: CharRef -> SwordRef -> Game (Maybe (CharRef, SwordRef))
chaseSword cr sr = do
  cs <- getCS cr
  if elem sr $ csSwords cs
    then return $ Just (cr, sr)
    else do
      tr <- liftM gsSwordTransfer getGS
      case M.lookup (cr, sr) tr of
        Just (cr', sr') -> chaseSword cr' sr'
        Nothing -> return Nothing

attackThenUnuse :: DamageFormula -> Effect
attackThenUnuse df = EAtk [simpleRE $ simpleAttackPrim df, unuseRE]
  where
    unuseRE = simpleRE $ targetNone $ \ec -> delayAC $ \_ -> do
      mp <- chaseSword (ecOwner ec) (ecThisSword ec)
      case mp of
        Nothing -> return ()
        Just (owner, ps) -> modifyCS owner $ \cs -> cs{ csSwordState = M.adjust(\ss -> ss{ ssUsed = False }) ps $ csSwordState cs }

silentAttack :: (CharRef -> CharRef -> Game ()) -> Effect
silentAttack f = EAtk [re]
  where
    re = simpleRE $ targetEnemy $ \y ec -> delayAC $ \_ -> f (ecOwner ec) y

flashfireAttack :: Effect
flashfireAttack = EAtk [simpleRE $ targetEnemy $ \tgt _ -> fire tgt, simpleRE $ targetNone $ \ec -> fire (ecOwner ec)]
  where
    fire x = delayAC $ \_ -> do
      mes $ show x ++ "の剣が全て炸撃剣になった"
      modifyCS x $ \cs -> cs { csSwordState = M.map tofc $ csSwordState cs }
    tofc ss = changeSwordType (initialSS firecracker) ss

changeSwordType :: SwordState -> SwordState -> SwordState
changeSwordType st ss = st { ssUsed = ssUsed ss }

handleDriveAndGuard :: Bool -> CharRef -> CharRef -> Integer -> Game Integer
handleDriveAndGuard penetrate from to n = do
  n' <- handleDrive from n
  if penetrate
    then return n'
    else handleGuard to n'

atkDamage :: Bool -> CharRef -> CharRef -> Integer -> Game ()
atkDamage mp from to dmg
  | dmg <= 0 = mes $ show to ++ "はダメージを受けなかった"
  | mp = mpDamage dmg to
  | otherwise = damage "攻撃" dmg to

mpDamage :: Integer -> CharRef -> Game ()
mpDamage n x
  | n <= 0 = return ()
  | otherwise = do
    cs <- getCS x
    let mp = csMp cs
    mes $ show x ++ "は" ++ show n ++ "点のMPダメージを受けた。MP: " ++ show mp ++ " -> " ++ show (mp - n)
    setCS x cs{ csMp = mp - n }

handleDrive :: CharRef -> Integer -> Game Integer
handleDrive x dmg = do
  cs <- getCS x
  let so = csStatusOther cs
  case soDrive so of
    0 -> return dmg
    n -> do
      let 
        atk = csAtk cs
        (newatk, newdmg)
          | atk > 0 = (atk - 1, dmg + 3)
          | otherwise = (atk, dmg)
      mes $ show x ++ "のドライブが発動した。攻: " ++ show atk ++ " -> " ++ show newatk ++ "、ダメージ: " ++ show dmg ++ " -> " ++ show newdmg
      setCS x cs{ csStatusOther = so{ soDrive = n - 1 }, csAtk = newatk }
      return newdmg

handleGuard :: CharRef -> Integer -> Game Integer
handleGuard x dmg = do
  cs <- getCS x
  case B.pop $ csStatusGuard cs of
    Nothing -> return dmg
    Just (g, rest) -> do
      mes $ show x ++ "の" ++ show g ++ "が発動した。" ++ extraMsg ++ "ダメージ: " ++ show dmg ++ " -> " ++ show newdmg
      setCS x cs'{ csStatusGuard = rest }
      return newdmg
      where
        (newdmg, extraMsg, cs') = case g of
          Barrier -> (0, "", cs)
          Cusion -> (1, "", cs)
          Flip
            | def > 0 -> (dmg - 3, "防: " ++ show def ++ " -> " ++ show newdef ++ "、", cs{ csDef = newdef })
            | otherwise -> (dmg, "", cs)
            where
              def = csDef cs
              newdef = def - 1
          Shield -> (div dmg 2, "", cs)
          Guard -> (div (dmg+1) 2, "", cs)

simpleAC :: (GameState -> Game ()) -> ActionChain
simpleAC f = AC $ \gs -> f gs >> return ([], Nothing)

delayAC :: (GameState -> Game ()) -> ActionChain
delayAC f = AC $ \gs -> return ([], Just $ simpleAC f)

simpleRE :: PrimEffect -> RichEffect
simpleRE pe = RE{ rePrim = pe, reRepeat = 1 }

targetAlly :: (CharRef -> EffectContext -> ActionChain) -> PrimEffect
targetAlly gen = TargetAlly{ taAll = False, taGen = gen }

targetEnemy :: (CharRef -> EffectContext -> ActionChain) -> PrimEffect
targetEnemy gen = TargetEnemy{ teDefendable = True, teAll = False, teGen = gen }

targetNone :: (EffectContext -> ActionChain) -> PrimEffect
targetNone gen = TargetNone{ tnGen = gen }

gain :: (Show a) => String -> (CharacterState -> a) -> (CharacterState -> CharacterState) -> Effect
gain name val adder = aux $ doGain name val adder

doGain :: (Show a) => String -> (CharacterState -> a) -> (CharacterState -> CharacterState) -> GameState -> EffectContext -> CharRef -> Game ()
doGain name val adder = \gs ec y -> do
  cs <- getCS y
  let cs' = adder cs
  mes $ show y ++ "は" ++ name ++ "を得た。" ++ show (val cs) ++ " -> " ++ show (val cs')
  setCS y cs'

doubleStatus :: CharRef -> Game ()
doubleStatus x = do
  cs <- getCS x
  let sss@[maxHp, hp, mp, atk, def, spd] = map ($ cs) [csMaxHp, csHp, csMp, csAtk, csDef, csSpd]
  let sss'@[maxHp', hp', mp', atk', def', spd'] = map (2*) sss
  mes $ show x ++ "の能力が倍になった。" ++ intercalate "、" (zipWith3 fmt sss sss' ["最大HP", "HP", "MP", "攻撃力", "防御力", "素早さ"])
  setCS x cs{ csMaxHp = maxHp', csHp = hp', csMp = mp', csAtk = atk', csDef = def', csSpd = spd' }
  where
    fmt from to name = name ++ ": " ++ show from ++ " -> " ++ show to

doubleAttack :: CharRef -> Game ()
doubleAttack x = do
  cs <- getCS x
  let old = csAtk cs
  let new = 2 * old
  mes $ show x ++ "の攻撃力が倍になった。" ++ show old ++ " -> " ++ show new
  setCS x cs{ csAtk = new }

putGuard :: GuardStatus -> Effect
putGuard gs = gain (show gs) (B.count gs . csStatusGuard) $ \cs -> cs{ csStatusGuard = B.insert gs $ csStatusGuard cs }

putRevive :: ReviveStatus -> Effect
putRevive rs = aux $ \gs ec y -> do
  doGain (show rs) (B.count rs . csStatusRevive) ( \cs -> cs{ csStatusRevive = B.insert rs $ csStatusRevive cs }) gs ec y
  cs <- getCS y
  when (csDead cs) $ do
    rev <- invokeRevive y
    when rev $ revive y

putDefence :: DefenceStatus -> Effect
putDefence gs = gain (show gs) (B.count gs . csStatusDefence) $ \cs -> cs{ csStatusDefence = B.insert gs $ csStatusDefence cs }

putOther :: String -> (StatusOther -> Int) -> (StatusOther -> StatusOther) -> Effect
putOther name get put = gain name (get . csStatusOther) $ \cs -> cs{ csStatusOther = put $ csStatusOther cs }

putOtherOne :: String -> (StatusOther -> Bool) -> (StatusOther -> StatusOther) -> Effect
putOtherOne name get put = gain name (boolToInt . get . csStatusOther) $ \cs -> cs{ csStatusOther = put $ csStatusOther cs }
  where
    boolToInt b = if b then 1 else 0

putProvoke :: Effect
putProvoke = putOtherOne "挑発" soProvoke $ \so -> so{ soProvoke = True }

aux :: (GameState -> EffectContext -> CharRef -> Game ()) -> Effect
aux f = EAux [re]
  where
    re = simpleRE $ targetAlly $ \y ec -> simpleAC $ \gs -> f gs ec y

ctl, ctlAll :: String -> ((CharacterState, Sword) -> Bool) -> (SwordState -> Maybe SwordState) -> Effect
ctl = ctlGeneral False
ctlAll = ctlGeneral True

ctlGeneral :: Bool -> String -> ((CharacterState, Sword) -> Bool) -> (SwordState -> Maybe SwordState) -> Effect
ctlGeneral allp desc cond trans = ECtl allp fn
  where
    fn gs ec = \ss -> let
      sw = ssSword ss
      in if cond (cs, sw)
        then (trans ss, [show owner ++ "の" ++ show ss ++ "は" ++ desc])
        else (Just ss, [])
      where
        cs = gsChars gs CR.! owner
        owner = ecOwner ec

ctlResonance :: Effect
ctlResonance = ECtl True fn
  where
    fn gs ec = case nextSwordAsSubtarget gs ec of
      Nothing -> \ss -> (Just ss, [])
      Just to -> \ss -> let
        msg = show (ecOwner ec) ++ "の" ++ show ss ++ "は" ++ show to ++ "になった"
        in (Just $ changeSwordType to ss, [msg])

bld :: Character -> Effect
bld ch = EBld $ \gs ec -> ch

primaryChar :: (TeamRef -> TeamRef) -> GameState -> EffectContext -> Character
primaryChar side gs ec = ciOriginal $ csInfo $ (css CR.!) $ teamcase primary0 primary1 $ side $ csSide $ css CR.! ecOwner ec
  where
    css = gsChars gs

defaultAttack :: Effect
defaultAttack = attack $ ATK - DEF

recoverAll :: CharRef -> CharRef -> Game ()
recoverAll x y = do
  cs <- getCS y
  mes $ show y ++ "は全回復した。" ++ hpChange (csHp cs) (csMaxHp cs)
  setHP y (csMaxHp cs)

destroySword :: GameState -> CharRef -> CharRef -> Game ()
destroySword gs x y = case csSwords $ gsChars gs CR.! y of
  [] -> return ()
  target:_ -> do
    cs <- getCS y
    mes $ show y ++ "の腰骨が折れた"
    mes $ show y ++ "の" ++ (show $ ssSword $ csSwordState cs M.! target) ++ "は破壊された"
    setCS y $ deleteSword target cs

stealSword :: GameState -> Integer -> CharRef -> CharRef -> Game [CharRef]
stealSword gs n x y = do
  when (n > 0) $ case csSwords $ gsChars gs CR.! y of
    [] -> return ()
    target:_ -> do
      cs <- getCS y
      let ss = csSwordState cs M.! target
      mes $ show x ++ "は" ++ show y ++ "の" ++ show ss ++ "を盗んだ"
      setCS y $ deleteSword target cs
      sn <- addSword x ss
      registerSwordTransfer y target x sn
  return []

pressurize :: GameState -> CharRef -> CharRef -> Game ()
pressurize gs x y = case filter (not . ssUsed . (csSwordState oldCS M.!)) $ csSwords oldCS of
  [] -> return ()
  target:_ -> do
    cs <- getCS y
    mes $ show y ++ "の" ++ show (ssSword $ csSwordState cs M.! target) ++ "は使用済みになった"
    setCS y cs{ csSwordState = M.adjust (\ss -> ss{ ssUsed = True }) target $ csSwordState cs }
  where
    oldCS = gsChars gs CR.! y

reverseSwords :: GameState -> CharRef -> CharRef -> Game ()
reverseSwords gs x y = do
  mes $ show y ++ "の剣の順序は逆順になった"
  modifyCS y $ \cs -> cs{ csSwords = reverse $ csSwords cs }

gainMp :: Integer -> Effect
gainMp n = gain "MP" csMp $ \v -> v{ csMp = csMp v + n }

gainProspect :: Effect
gainProspect = gain "最大HPとHPとスリープ" select add
  where
    select cs = (csMaxHp cs, csHp cs, sbSleep $ csStatusBad cs)
    add v = v{ csMaxHp = csMaxHp v + 10, csHp = csHp v + 10, csStatusBad = sb{ sbSleep = sbSleep sb + 1 } }
      where
        sb = csStatusBad v

resetAttack, resetDefence, resetSpeed :: CharRef -> CharRef -> Game ()
resetAttack = resetStatus "攻撃力" chAtk csAtk $ \cs n -> cs{ csAtk = n }
resetDefence = resetStatus "防御力" chDef csDef $ \cs n -> cs{ csDef = n }
resetSpeed = resetStatus "素早さ" chSpd csSpd $ \cs n -> cs{ csSpd = n }

resetStatus :: String -> (Character -> Integer) -> (CharacterState -> Integer) -> (CharacterState -> Integer -> CharacterState) -> CharRef -> CharRef -> Game ()
resetStatus name get gets set x y = do
  cs <- getCS y
  let newval = get $ ciOriginal $ csInfo cs
  mes $ show y ++ "の" ++ name ++ "が初期値に戻った。" ++ show (gets cs) ++ " -> " ++ show newval
  setCS y $ set cs newval

removeAllBad :: CharRef -> Game ()
removeAllBad x = do
  mes $ show x ++ "の悪い状態が全て除去された"
  modifyCS x $ \cs -> cs{ csStatusBad = noBadStatus }

nextSwordAsSubtarget :: GameState -> EffectContext -> Maybe SwordState
nextSwordAsSubtarget gs ec = fmap snd $ getECNext gs ec

enemySword :: GameState -> EffectContext -> Maybe SwordState
enemySword gs ec = fmap (csSwordState cs M.!) msr
  where
    enemy = head $ gsTeam (opponent $ csSide $ chars CR.! ecOwner ec) gs
    cs = chars CR.! enemy
    msr = case csSwords cs of
      sr:_ -> Just sr
      _ -> Nothing
    chars = gsChars gs

isAtk, isAux :: Sword -> Bool
isAtk sw = case swEffect sw of
  EAtk{} -> True
  _ -> False
isAux sw = case swEffect sw of
  EAux{} -> True
  _ -> False

recover :: CharRef -> Integer -> Game ()
recover cr n = do
  cs <- getCS cr
  let hp = csHp cs
  let newhp = max hp $ min (csMaxHp cs) $ hp + n
  let d = newhp - hp
  when (d /= 0) $ do
    mes $ show cr ++ "は" ++ show d ++ "点回復した。" ++ hpChange hp newhp
    setHP cr newhp

recoverSelf :: Integer -> CharRef -> CharRef -> Game [CharRef]
recoverSelf n self _ = do
  recover self n
  return []

sw :: String -> String -> Effect -> Sword
sw name short eff = Sw
  { swName = name
  , swNamespace = M100
  , swShortName = short
  , swEffect = eff
  , swTurn = 1
  , swLimited = False
  , swPeriod = Nothing
  , swMP = Nothing
  , swReactsToDanger = False
  , swReactsToIllness = False
  }

swa :: String -> String -> Effect -> Sword
swa name short eff = (sw name short eff) { swNamespace = M2011a }

swn :: Int -> String -> String -> Effect -> Sword
swn n name short eff = (sw (name ++ sn) (short ++ sn) eff) { swTurn = n, swLimited = True }
  where
    sn = show n

swpn :: Int -> String -> String -> Effect -> Sword
swpn n name short eff = (sw (name ++ "." ++ sn) (short ++ "." ++ sn) eff) { swTurn = 0, swPeriod = Just n }
  where
    sn = show n

data DamageFormula
  = ATK
  | DEF
  | HP
  | TURN
  | SelfMaxHP
  | SelfHP
  | DFInt Int
  | DFOp1 DFOp1 DamageFormula
  | DFOp2 DFOp2 DamageFormula DamageFormula
  deriving (Eq, Show)

data DFOp1 = DFOpNegate | DFOpAbs | DFOpSignum
  deriving (Eq, Show)
data DFOp2 = DFOpAdd | DFOpMul | DFOpDiv | DFOpMod | DFOpPow
  deriving (Eq, Show)

instance Num DamageFormula where
  fromInteger n = DFInt $ fromInteger n
  a + b = DFOp2 DFOpAdd a b
  a * b = DFOp2 DFOpMul a b
  negate = DFOp1 DFOpNegate
  abs = DFOp1 DFOpAbs
  signum = DFOp1 DFOpSignum

dfdiv, dfmod, dfpow :: DamageFormula -> DamageFormula -> DamageFormula
dfdiv = DFOp2 DFOpDiv
dfmod = DFOp2 DFOpMod
dfpow = DFOp2 DFOpPow

evalDamageFormula :: Int -> CharacterState -> CharacterState -> DamageFormula -> Integer
evalDamageFormula turn self tgt f = case f of
  ATK -> csAtk self
  DEF -> csDef tgt
  HP -> csHp tgt
  SelfMaxHP -> csMaxHp self
  SelfHP -> csHp self
  TURN -> toInteger turn
  DFInt n -> toInteger n
  DFOp1 op1 g -> evalDFOp1 op1 (evalDamageFormula turn self tgt g)
  DFOp2 op2 g h -> evalDFOp2 op2 (evalDamageFormula turn self tgt g) (evalDamageFormula turn self tgt h)

evalDamageFormulaFromGS :: GameState -> CharRef -> CharRef -> DamageFormula -> Integer
evalDamageFormulaFromGS gs x y = evalDamageFormula (gsTurnCount gs) (gsChars gs CR.! x) (gsChars gs CR.! y)

evalDFOp1 :: DFOp1 -> Integer -> Integer
evalDFOp1 o = case o of
  DFOpNegate -> negate
  DFOpAbs -> abs
  DFOpSignum -> signum

evalDFOp2 :: DFOp2 -> Integer -> Integer -> Integer
evalDFOp2 o = case o of
  DFOpAdd -> (+)
  DFOpMul -> (*)
  DFOpDiv -> div
  DFOpMod -> mod
  DFOpPow -> (^)

putAbortion :: GameState -> CharRef -> CharRef -> Game ()
putAbortion = putBad "流産" sbAbortion $ \b -> b{ sbAbortion = sbAbortion b + 1 }

putEarly :: Effect
putEarly = ctl "早期起動を得た" (const True) $ \ss -> Just ss{ ssEarly = ssEarly ss + 1 }

makeAllUnused :: Effect
makeAllUnused = ctlAll "未使用になった" (const True) $ \ss -> Just $ ss{ ssUsed = False }

putPush :: Effect
putPush = putOther "プッシュ" soPush $ \so -> so{ soPush = soPush so + 1 }
