# include "internal.h"
# include <sstream>

using namespace m100sim;

namespace
{

void pr_real(FILE *out, const char *title, int count)
{
  switch(count)
  {
  case 1:
    fprintf(out, " <%s>", title);
    return;
  case 2:
    fprintf(out, " <%s> <%s>", title, title);
    return;
  default:
    fprintf(out, " <%s>*%d", title, count);
  }
}

inline void pr(FILE *out, const char *title, int count)
{
  if(count) pr_real(out, title, count);
}

template <typename T>
void prms(FILE *out, const multiset<T> &ms)
{
  typename multiset<T>::iterator it = ms.begin();
  while(it != ms.end())
  {
    T cur = *it;
    int c = 0;
    do
    { 
      ++it;
      c++;
    }
    while(it != ms.end() && *it == cur);
    pr(out, name(cur), c);
  }
}

void bring_death(gstate &gs, logger &lg)
{
  for(int i = 0; i < 2; i++)
    for(list<cstate>::iterator ci = gs.teams[i].chs.begin(); ci != gs.teams[i].chs.end(); ++ci)
      if(ci->hp <= 0)
        really_kill(gs, &*ci, lg);
}

void handle_period(gstate &gs, logger &lg)
{
  for(int i = 0; i < 2; i++)
    for(list<cstate>::iterator ci = gs.teams[i].chs.begin(); ci != gs.teams[i].chs.end(); ++ci)
    {
      list<sstate> to_remove;
      for(sref s = ci->swords.begin(); s != ci->swords.end(); )
        if(s->info->period && *s->info->period < gs.turn)
        {
          sref next = s;
          ++next;
          to_remove.splice(to_remove.end(), ci->swords, s);
          s = next;
        }
        else
          ++s;
      if(!to_remove.empty())
        if(FILE *out = lg.start_direct())
        {
          bool first = true;
          fprintf(out, "%sの", ci->info.name.c_str());
          for(sref s = to_remove.begin(); s != to_remove.end(); ++s)
          {
            if(!first) fputs("、", out);
            print_sstate(out, *s);
            first = false;
          }
          fputs("はピリオドによって消滅した", out);
          lg.end_direct();
        }
    }
}

}

void m100sim::revive(gstate &gs, cstate *tgt, logger &lg)
{
  if(tgt->dead)
  {
    lg.putf("%sは復活した", tgt->info.name.c_str());
    tgt->dead = false;
  }
}

void m100sim::really_kill(gstate &gs, cstate *tgt, logger &lg)
{
  lg.putf("%sは死んだ", tgt->info.name.c_str());
  if(!invoke_revival(gs, tgt, lg))
  {
    tgt->dead = true;
    gs.someone_may_be_dead = true;
  }
}

bool m100sim::invoke_revival(gstate &gs, cstate *tgt, logger &lg)
{
  while(tgt->hp <= 0)
  {
    multiset<revival_status> &rs = tgt->revival;
    multiset<revival_status>::iterator fst = rs.begin();
    if(fst == rs.end())
      return false;
    else
    {
      int_t newhp;
      switch(*fst)
      {
      case life:
        newhp = 1;
        break;
      case soul:
        newhp = tgt->max_hp;
        break;
      }
      lg.put(boost::format("%1%が発動した。HP: %2% -> %3%")
          % *fst % tgt->hp % newhp);
      tgt->hp = newhp;
      rs.erase(fst);
    }
  }
  return true;
}

void m100sim::kill(gstate &gs, cstate *tgt, logger &lg)
{
  if(gs.turn != 0)
    really_kill(gs, tgt, lg);
}

void m100sim::set_hp(gstate &gs, cstate *tgt, intparam val, logger &lg)
{
  bool dead = tgt->hp <= 0;
  tgt->hp = val;
  if(!dead && val <= 0)
    kill(gs, tgt, lg);
  else if(dead && val > 0)
    revive(gs, tgt, lg);
}

void m100sim::damage(gstate &gs, const char *reason, cstate *tgt, intparam amount, logger &lg)
{
  if(amount <= 0) return;
  int_t newhp = checked_subtract(tgt->hp, amount);
  if(FILE *out = lg.start_direct())
  {
    fprintf(out, "%sは%sによって", tgt->info.name.c_str(), reason);
    print_int(out, amount);
    fputs("点のダメージを受けた。HP: ", out);
    print_int(out, tgt->hp);
    fputs(" -> ", out);
    print_int(out, newhp);
    lg.end_direct();
  }
  set_hp(gs, tgt, newhp, lg);
}

const char *m100sim::name(guard_status s)
{
  switch(s)
  {
  case barrier: return "バリア";
  case cusion: return "クッション";
  case flip: return "フリップ";
  case shield: return "シールド";
  case guard: return "ガード";
  }
}

const char *m100sim::name(defence_status s)
{
  switch(s)
  {
  case reflection: return "反射";
  case detection: return "見切り";
  }
}

const char *m100sim::name(revival_status s)
{
  switch(s)
  {
  case life: return "ライフ";
  case soul: return "ソウル";
  }
}

void m100sim::print_sstate(FILE *ost, const sstate &ss, bool show_usedness)
{
  if(ss.used && show_usedness)
    fputc('(', ost);
  fputs(ss.info->name, ost);
  for(int i = 0; i < ss.ench.early; i++) fputs("<早>", ost);
  for(int i = 0; i < -ss.ench.early; i++) fputs("<遅>", ost);
  for(int i = 0; i < ss.ench.stun; i++) fputs("<振>", ost);
  for(int i = 0; i < ss.ench.poison; i++) fputs("<与毒>", ost);
  if(ss.ench.scatter) fputs("<全>", ost);
  if(ss.ench.limited) fputs("<リミ>", ost);
  if(ss.used && show_usedness)
    fputc(')', ost);
}

void m100sim::print_cstate(FILE *out, const cstate &cs)
{
  fputs(cs.info.name.c_str(), out);
  fputc(' ', out);
  print_int(out, cs.hp);
  fputc('(', out);
  print_int(out, cs.max_hp);
  fputs(")/", out);
  print_int(out, cs.atk);
  fputc('/', out);
  print_int(out, cs.def);
  fputc('/', out);
  print_int(out, cs.spd);
  for(list<sstate>::const_iterator it = cs.swords.begin(); it != cs.swords.end(); ++it)
  {
    fputc(' ', out);
    print_sstate(out, *it);
  }
  if(cs.mp != 0)
  {
    fputs(" MP:", out);
    print_int(out, cs.mp);
  }
  if(cs.ap != 0)
  {
    fputs(" AP:", out);
    print_int(out, cs.ap);
  }
  if(cs.reserved_ap != 0)
  {
    fputs(" 予備AP:", out);
    print_int(out, cs.reserved_ap);
  }
  pr(out, "スタン", cs.bad.stun);
  pr(out, "スリープ", cs.bad.sleep);
  pr(out, "猛毒", cs.bad.bad_poison);
  pr(out, "毒", cs.bad.poison);
  pr(out, "流産", cs.bad.abortion);
  prms(out, cs.guard);
  prms(out, cs.defence);
  prms(out, cs.revival);
  pr(out, "プッシュ", cs.other.push);
  pr(out, "ヘヴンズドア", cs.other.heavens_door);
  pr(out, "クイック", cs.other.quick);
  pr(out, "スロー", cs.other.slow);
  pr(out, "ドライブ", cs.other.drive);
  pr(out, "挑発", cs.other.provocation);
}

void m100sim::increment_turn_count(gstate &gs, logger &lg)
{
  {
    indent_controller ic(lg, -lg.current_indent());
    lg.putf("<<< Turn %d >>>", gs.turn + 1);
  }
  if(gs.turn == 0)
    bring_death(gs, lg); /* The blessed 0th turn is over; now all swordsmen are mortal. */
  gs.turn++;
  handle_period(gs, lg);
}

void m100sim::register_to_action_order(gstate &gs, cstate *self)
{
  quickness q(self);
  gstate::action_order_type::iterator p = gs.action_order.find(q);
  if(p == gs.action_order.end())
    p = gs.action_order.insert(make_pair(q, cset())).first;
  p->second.insert(self);
  if(self->ap > 0 && (gs.move_next == gs.action_order.end() || gs.move_next->first < q))
    gs.move_next = p;
}

void m100sim::unregister_from_action_order(gstate &gs, cstate *self)
{
  gstate::action_order_type::iterator p = gs.action_order.find(quickness(self));
  p->second.erase(self);
  if(p->second.empty())
  {
    if(gs.move_next == p)
      ++gs.move_next;
    gs.action_order.erase(p);
  }
}

cstate *m100sim::gstate::add_character(const cinfo &c)
{
  teams[c.side].chs.push_front(cstate(c));
  cstate *self = &teams[c.side].chs.front();
  register_to_action_order(*this, self);
  return self;
}
