# include <stdio.h>
# include <unistd.h>
# include <stdlib.h>
# include <locale.h>
# include <fcntl.h>

# include "haskell_interface.h"
# include "match.h"

# ifdef USE_GMP_INTEGER
#  include "gmp.h"
# endif

static void fill_character(m100sim_character *out, const char *name, int hp, int atk, int def, int spd, const m100sim_sword_ref *swords, int nsword)
{
  out->name = name;
  out->hp = hp;
  out->atk = atk;
  out->def = def;
  out->spd = spd;
  out->swords = swords;
  out->n_swords = nsword;
}

# ifdef USE_GMP_INTEGER
static void *my_realloc(void *p, size_t oldsz, size_t newsz)
{
  return realloc(p, newsz);
}

static void my_free(void *p, size_t sz)
{
  free(p);
}
# endif

m100sim_match_result m100sim_match_for_haskell(int fd_out,
    const wchar_t *xname, int xhp, int xatk, int xdef, int xspd, const m100sim_sword_ref *xswords, int xn_sword,
    const wchar_t *yname, int yhp, int yatk, int ydef, int yspd, const m100sim_sword_ref *yswords, int yn_sword)
{
  m100sim_match_options opts;
  m100sim_match_result res;
  m100sim_character x, y;
  char xbuf[1000], ybuf[1000];
# ifdef USE_GMP_INTEGER
  void *(*p_alloc)(size_t);
  void *(*p_realloc)(void *, size_t, size_t);
  void (*p_free)(void *, size_t);
# endif

  setlocale(LC_CTYPE, "");

# ifdef USE_GMP_INTEGER
  mp_get_memory_functions(&p_alloc, &p_realloc, &p_free);
  mp_set_memory_functions(malloc, my_realloc, my_free);
# endif

  size_t rx, ry;
  if((rx = wcstombs(xbuf, xname, 1000)) >= 1000 || wcstombs(ybuf, yname, 1000) >= 1000)
  {
    int i;
    fprintf(stderr, "name(%d): ", (int)rx);
    for(i = 0; xname[i]; i++)
      fprintf(stderr, " %d", (int)xname[i]);
    fprintf(stderr, "\nm100sim haskell interface: failed to convert character names\n");
    return m100sim_aborted;
  }

  if(fd_out >= 0)
    fcntl(fd_out, F_SETFL, (long)fcntl(fd_out, F_GETFL) & ~O_NONBLOCK);
  opts.log_output = fd_out >= 0 ? fdopen(dup(fd_out), "a") : NULL;
# ifdef __MINGW32__
  opts.line_terminator = "\r\n";
# else
  opts.line_terminator = "\n";
# endif
  fill_character(&x, xbuf, xhp, xatk, xdef, xspd, xswords, xn_sword);
  fill_character(&y, ybuf, yhp, yatk, ydef, yspd, yswords, yn_sword);
  res = m100sim_match(opts, x, y);

# ifdef USE_GMP_INTEGER
  mp_set_memory_functions(p_alloc, p_realloc, p_free);
# endif
  if(opts.log_output)
    fclose(opts.log_output);
  if(fd_out >= 0)
    fcntl(fd_out, F_SETFL, (long)fcntl(fd_out, F_GETFL) | O_NONBLOCK);
  return res;
}
