# include <stdio.h>
# include <string.h>
# include <windows.h>

/* Filter to convert UTF-8 input to an encoding determined
 * by the current locale.
 * It also escapes each occurrence of 0x5c in the output.
 * Used for preprocessing source files before compiling them with mingw-gcc,
 * which assumes input files to be in UTF-8.
 */

# define BUFSIZE 1000

int preescape(const wchar_t *src, int len, wchar_t *dest)
{
  int i;
  int n = 0;
  for(i = 0; i < len; i++)
  {
    if(src[i] == L'\\')
      dest[n++] = L'\\';
    dest[n++] = src[i];
  }
  return n;
}

void print_escaping(const char *src, int len)
{
  int escaped = 0;
  int i;
  for(i = 0; i < len; i++)
  {
    if(escaped)
    {
      if(src[i] != '\\') /* accidental 0x0c */
        printf("\\\\%c", src[i]);
      else
        putchar('\\');
      escaped = 0;
    }
    else
      if(src[i] != '\\')
        putchar(src[i]);
      else
        escaped = 1;
  }
  if(escaped)
  {
    fprintf(stderr, "line improperly terminated\n");
    exit(1);
  }
}

int main(void)
{
  char ibuf[BUFSIZE];
  while(fgets(ibuf, BUFSIZE, stdin))
  {
    wchar_t wbuf0[BUFSIZE], wbuf1[BUFSIZE*2];
    char obuf[BUFSIZE*4];
    int ilen, wlen0, wlen1, olen;
    BOOL invalid;
    ilen = strlen(ibuf);
    if(ibuf[ilen-1] != '\n')
    {
      fprintf(stderr, "line too long\n");
      return 1;
    }
    wlen0 = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, ibuf, ilen, wbuf0, BUFSIZE);
    if(wlen0 == 0)
    {
      fprintf(stderr, "conversion to wchar_t failed\n");
      return 1;
    }
    wlen1 = preescape(wbuf0, wlen0, wbuf1);
    olen = WideCharToMultiByte(CP_ACP, 0, wbuf1, wlen1, obuf, BUFSIZE*4, NULL, &invalid);
    if(olen == 0 || invalid)
    {
      fprintf(stderr, "conversion from wchar_t failed\n");
      return 1;
    }
    print_escaping(obuf, olen);
  }
  return 0;
}
