{-# OPTIONS_GHC -fglasgow-exts #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

  Block port implementation on top of 'System.IO.File.File'.
-}
module System.IO.SSC.FilePort
  ( FilePort
  , filePort
  ) where

import Control.Concurrent
import Control.Monad
import Data.Maybe
import Data.Word
import Data.Yesno
import Foreign.Ptr
import Foreign.Marshal.Array
import System.IO(SeekMode(..))
import System.IO.File
import System.IO.SSC.Classes
import System.IO.SSC.Internal.Fd
import System.IO.SSC.Internal.Debug
import System.IO.SSC.Internal.BasicPort

newtype FilePort r w = FP (BasicPort (File r w))
type FilePortR = FilePort YesTag NoTag
type FilePortW = FilePort NoTag YesTag
type FilePortRW = FilePort YesTag YesTag

filePort :: File r w -> IO (FilePort r w)
filePort f = liftM FP $ basicPort f

instance Port (FilePort r w) where
  rawClose (FP bp) = bpClose bp fClose
  rawIsClosed (FP bp) = bpIsClosed bp
  rawFlush _ = return ()

instance Destination (File r w) (FilePort r w) where
  rawDest (FP bp) = return (bpBase bp)

instance Lockable (FilePort r w) where
  sLock (FP bp) = bpLock bp
  sTryLock (FP bp) = bpTryLock bp
  sUnsafeUnlock (FP bp) = bpUnsafeUnlock bp

instance Discardable (FilePort r w) where
  rawDiscard _ = return ()

instance Seekable Integer (FilePort r w) where
  rawGetPos (FP bp) = bpEnsureOpen bp >> fTell (bpBase bp)
  rawSetPos (FP bp) pos = bpEnsureOpen bp >> fSeek (bpBase bp) AbsoluteSeek pos

instance RelativeSeekable Integer (FilePort r w) where
  rawRelativeSeek (FP bp) pos = bpEnsureOpen bp >> fSeek (bpBase bp) RelativeSeek pos

instance SizedInput Integer (FilePort r w) where
  rawInputEnd (FP bp) = bpEnsureOpen bp >> fFileSize (bpBase bp)
  rawSeekFromEnd (FP bp) pos = bpEnsureOpen bp >> fSeek (bpBase bp) SeekFromEnd pos

instance IBlock Word8 (FilePort YesTag w) where
  rawGetBufB (FP bp) bmode buf size = do
    bpEnsureOpen bp
    liftM fromJust $ fWithFd (bpBase bp) $ \fd -> fdGetBufB fd bmode buf size

instance OBlock Word8 (FilePort r YesTag) where
  rawPutBufB (FP bp) bmode buf size = do
    bpEnsureOpen bp
    msg $ "+rawPutBufB/FilePort size=" ++ show size
    r <- liftM fromJust $ fWithFd (bpBase bp) $ \fd -> fdPutBufB fd bmode buf size
    msg $ "-sPufBufB/FilePort r=" ++ show r
    return r
