{-# OPTIONS_GHC -fglasgow-exts #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   Input hook transformer provides a hook to install a hander, which
   is called whenever the input is requested. The idea is that it can
   be used to flush stdout every time before waiting for input at
   stdin, for example.
-}
module System.IO.SSC.InputHook
  ( InputHook
  , iInputHook
  , ioInputHook
  ) where

import System.IO.SSC.Classes

data InputHook d pos a ba ma b = IH b (IO ())

inputHook handler b = return (IH b handler)

iInputHook :: IO () -> b -> IO (InputHook d pos a ba ma b)
iInputHook = inputHook

ioInputHook :: IO () -> b -> IO (InputHook d pos a ba ma b)
ioInputHook = inputHook

instance (Port b) => Port (InputHook d pos a ba ma b) where
  rawClose (IH b _) = rawClose b
  rawIsClosed (IH b _) = rawIsClosed b
  rawFlush (IH b _) = rawFlush b

instance (Destination d b) => Destination d (InputHook d pos a ba ma b) where
  rawDest (IH b _) = rawDest b

instance (Lockable b) => Lockable (InputHook d pos a ba ma b) where
  sLock (IH b _) = sLock b
  sTryLock (IH b _) = sTryLock b
  sUnsafeUnlock (IH b _) = sUnsafeUnlock b

instance (Discardable b) => Discardable (InputHook d pos a ba ma b) where
  rawDiscard (IH b _) = rawDiscard b

instance (Seekable pos b) => Seekable pos (InputHook d pos a ba ma b) where
  rawGetPos (IH b _) = rawGetPos b
  rawSetPos (IH b _) = rawSetPos b

instance (RelativeSeekable pos b) => RelativeSeekable pos (InputHook d pos a ba ma b) where
  rawRelativeSeek (IH b _) = rawRelativeSeek b

instance (SizedInput pos b) => SizedInput pos (InputHook d pos a ba ma b) where
  rawInputEnd (IH b _) = rawInputEnd b
  rawSeekFromEnd (IH b _) = rawSeekFromEnd b

instance (SizedOutput pos b) => SizedOutput pos (InputHook d pos a ba ma b) where
  rawOutputEnd (IH b _) = rawOutputEnd b

instance (IStream a b) => IStream a (InputHook d pos a ba ma b) where
  rawGet (IH b a) = a >> rawGet b
  rawGets (IH b a) n = a >> rawGets b n
  rawGetContents (IH b a) = a >> rawGetContents b

instance (RetrievableStream a b) => RetrievableStream a (InputHook d pos a ba ma b) where
  rawGetUntil (IH b a) end = a >> rawGetUntil b end

instance (OStream a b) => OStream a (InputHook d pos a ba ma b) where
  rawPut (IH b _) x = rawPut b x
  rawPuts (IH b _) x = rawPuts b x

instance (IBlock ba b) => IBlock ba (InputHook d pos a ba ma b) where
  rawGetBufB (IH b a) bmode buf size = a >> rawGetBufB b bmode buf size

instance (OBlock ba b) => OBlock ba (InputHook d pos a ba ma b) where
  rawPutBufB (IH b _) = rawPutBufB b

instance (IMemory ma b) => IMemory ma (InputHook d pos a ba ma b) where
  rawReadMem (IH b _) = rawReadMem b
  rawEndRead (IH b _) = rawEndRead b
  rawFillMemB (IH b a) bmode = a >> rawFillMemB b bmode

instance (OMemory ma b) => OMemory ma (InputHook d pos a ba ma b) where
  rawWriteMem (IH b _) = rawWriteMem b
  rawEndWrite (IH b _) = rawEndWrite b
  rawFlushMemB (IH b _) = rawFlushMemB b
