-- #hide
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental
   
   This module contains some definitions convinient for defining
   memory IO ports.
-}
module System.IO.SSC.Internal.RawBuf where

import Control.Concurrent.MVar
import Control.Monad
import Data.IORef
import Foreign.ForeignPtr
import Foreign.Storable
import Foreign.Ptr
import Foreign.Marshal.Array(advancePtr)

import System.IO.SSC.Internal.Ptr
import System.IO.SSC.Classes

data RawBuf a extra = RB
  { rbFptr :: {-# UNPACK #-} !(ForeignPtr a)
  , rbBase :: {-# UNPACK #-} !(Ptr a)
  , rbEndBuf :: {-# UNPACK #-} !(Ptr a)
  , rbBegin :: {-# UNPACK #-} !(IORef (Ptr a))
  , rbEnd :: {-# UNPACK #-} !(IORef (Ptr a))
  , rbEOI :: IORef Bool
  , rbExtraObj :: {-# UNPACK #-} !extra
  }

newRawBuf :: (Storable a) => Int -> e -> IO (RawBuf a e)
newRawBuf size extra = do
  buf <- mallocForeignPtrArray size
  let base = unsafeForeignPtrToPtr buf
  let endBuf = advancePtr base size
  begin <- newIORef base
  end <- newIORef base
--  locker <- sGetLocker extra
  eoi <- newIORef False
  return $! RB
    { rbFptr = buf
    , rbBase = base
    , rbEndBuf = endBuf
    , rbBegin = begin
    , rbEnd = end
    , rbExtraObj = extra
--    , rbLockerObj = locker
    , rbEOI = eoi
    }

rbEmpty :: RawBuf a e -> IO Bool
rbEmpty RB{ rbBegin = begin, rbEnd = end } = liftM2 (==) (readIORef begin) (readIORef end)

rbUsed :: (Storable a) => RawBuf a e -> IO Int
rbUsed RB{ rbBegin = begin, rbEnd = end } = liftM2 distancePtr (readIORef end) (readIORef begin)

rbReadMem :: (Storable a) => RawBuf a e -> IO (Ptr a, Ptr a)
rbReadMem RB{ rbBegin = begin, rbEnd = end } = liftM2 (,) (readIORef begin) (readIORef end)

rbEndRead :: (Storable a) => RawBuf a e -> Ptr a -> IO ()
rbEndRead RB{ rbBegin = begin } endp = writeIORef begin endp

rbWriteMem :: (Storable a) => RawBuf a e -> IO (Ptr a, Ptr a)
rbWriteMem RB{ rbEnd = end, rbEndBuf = endb } = do
  vend <- readIORef end
  return (vend, endb)

rbEndWrite :: (Storable a) => RawBuf a e -> Ptr a -> IO ()
rbEndWrite RB{ rbEnd = end } endp = writeIORef end endp

rbReset :: RawBuf a e -> IO ()
rbReset rb@RB{ rbEOI = reoi } = do
  rbResetPtr rb
  writeIORef reoi False

rbResetPtr :: RawBuf a e -> IO ()
rbResetPtr RB{ rbBase = base, rbBegin = begin, rbEnd = end } = do
  writeIORef begin base
  writeIORef end base

rbTouch :: RawBuf a e -> IO ()
rbTouch RB{ rbFptr = fptr } = touchForeignPtr fptr

rbExtra :: RawBuf a e -> e
rbExtra RB{ rbExtraObj = obj } = obj

--rbLocker :: RawBuf a e -> MVar ()
--rbLocker RB{ rbLockerObj = locker } = locker
